# -*- coding: utf-8 -*-
r"""Scattering methods for Materials class

"""


class HKLGenerator(object):
    r"""Methods for generating HKL reflections

    """
    def generate_hkl_positions(self):
        r"""Generates hkl positions where scattering is allowed
        """
        pass

    def find_equivalent_positions(self):
        r"""Eliminates equivalent hkl positions
        """
        pass

    def apply_scattering_rules(self):
        r"""Applies scattering rules for the space group
        """
        pass

    def find_site_multiplicity(self):
        r"""Finds the multiplicity of the hkl position
        """
        pass
