# -*- coding: utf-8 -*-
r"""Polarization Tools

Get correction of polarized cross sections
Get He3 decay time

"""


def polarization_correction(data_pp, data_pm, data_mp, data_mm):
    r"""
    """
    pass
